! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
! catcorr_minor.f90  cgp 2011 July 31
! Routines with no need to access the catcorr_mod storage or structure definitions
module catcorr_minor
use types
use constants	! in ssclib, for Pi and TwoPi
implicit none
!integer, parameter :: double = selected_real_kind(10, 40)
private
public :: cona2v, congen, conmin, conmxm, conv2a, convcp, convrt, coord_trans, &
          dinvnorm, find_last_record, erfcc, gcdist, &
          ! getFreeIoUnit, mjd, &	! getFreeIoUnit() is already defined in ssclib/ssc_misc
          mjd, &
          poisson_ge, poisson_level, poisson_count, poisson_prob, &
          ran_gauss, rot_shift, rotate, splitString, stats
interface ahav
   module procedure ahavr, ahavd
end interface
interface gcdist
   module procedure gcdistd, gcdistr
end interface
interface hav
   module procedure havr, havd
end interface
! poisson* routines also in file poisson_subs.f90
! integer, parameter  :: dp = selected_real_kind(14,50)
! real(double), parameter :: pi = 3.14159265358979323846_double, twopi = 2.0_double * pi, &
real(double), parameter :: &
                       halfpi = 0.5_double * pi, &
                       rtod = 180.0_double/pi,            dtor = pi/180.0_double, &
                       rtoas = 3600.0_double*180.0_double/pi, astor = 1.0_double/rtoas
CONTAINS
!----------------------------------------------------------------------------
! ahav: inverse haversine, result in radians, real and double precision forms
real(double) function ahavd(x)         
real(double), intent(in) :: x
  ahavd = 2.0d0 * asin(sqrt(x))
end function ahavd
!----------------------------------------------------------------------------
real function ahavr(x)          ! inverse haversign, result in radians.
real, intent(in) :: x
  ahavr = 2.0 * asin(sqrt(x))
end function ahavr
!------------------------------------------------------------------------------
!+cona2v  Converts celestial position to a 3-vector
SUBROUTINE cona2v(az, el, v)
real(double), intent(in)  :: az    ! R.A. (or longitude or azimuth), radians.
real(double), intent(in)  :: el    ! Dec (or latitude or elevation), radians
real(double), intent(out) :: v(3)  ! 3-vector of unit length
!-
real(double) :: cel
!
cel = COS(el)
v(1) = COS(az) * cel
v(2) = SIN(az) * cel
v(3) = SIN(el)
END SUBROUTINE cona2v
!-------------------------------------------------------------------------------
!+CONGEN     Generates coordinate Conversion matrix from two 3-vectors
SUBROUTINE congen(vz, vx, ctos)
real(double), intent(in) :: vz(3)       ! Vector of pole of new frame in old frame
real(double), intent(in) :: vx(3)       ! Vector of reference point on equator 
real(double), intent(out) :: ctos(3, 3)  ! Unit matrix for converting from old to new frame
!-
ctos(:,3) = vz
CALL convcp(vz, vx, ctos(:,2))
CALL convcp(ctos(:,2), vz, ctos)
END SUBROUTINE congen
!-------------------------------------------------------------------------------
!+CONMIN     Inverts a coordinate conversion matrix (by transpose)
SUBROUTINE conmin(ctos, stoc)
real(double), INTENT(IN)   :: ctos(3, 3)  ! Conversion matrix for frame A to B
real(double), INTENT(OUT)  :: stoc(3, 3)  ! Conversion matrix for frame B to A
!-
stoc = transpose(ctos)
END SUBROUTINE conmin
!-------------------------------------------------------------------------------
!+CONMXM     Multiplies two 3x3 matrices [C]=[A].[B]
SUBROUTINE conmxm(a, b, c)
real(double), INTENT(IN)  :: a(3, 3)
real(double), INTENT(IN)  :: b(3, 3)
real(double), INTENT(OUT) :: c(3, 3)  ! Matrix product: [C] = [A].[B]
integer :: i, j, k
!
DO j = 1, 3
  DO i = 1, 3
    c(i, j) = 0.0
    DO k = 1, 3
      c(i, j) = c(i, j) + a(i, k) * b(k, j)
    END DO
  END DO
END DO
END SUBROUTINE conmxm
!-------------------------------------------------------------------------------
!+CONV2A     Converts 3-vector to spherical polars, radians.
SUBROUTINE conv2a(v, az, el)
real(double), intent(in)  :: v(3)    ! Three-vector
real(double), intent(out) :: az, el  ! Azimuth and elevation (or RA, dec), radians
!-
az = ATAN2(v(2), v(1))
IF(az < 0.0_double) az = TWOPI + az
el = ASIN(v(3))
END SUBROUTINE conv2a
!-------------------------------------------------------------------------------
!+CONVCP     Takes cross-product of two 3-vectors
SUBROUTINE convcp(va, vb, vc)
real(double), INTENT(IN)  :: va(3)
real(double), INTENT(IN)  :: vb(3)
real(double), INTENT(OUT) :: vc(3)  ! Vector cross-product, i.e. VC = VA X VB
!-
real(double) :: vnorm
!
vc(1) = va(2)*vb(3) - va(3)*vb(2)
vc(2) = va(3)*vb(1) - va(1)*vb(3)
vc(3) = va(1)*vb(2) - va(2)*vb(1)
vnorm = SQRT(vc(1)**2 + vc(2)**2 + vc(3)**2)
vc = vc / vnorm
END SUBROUTINE convcp
!-------------------------------------------------------------------------------
!+CONVRT   Converts spherical-polar coordinates from one frame to another
SUBROUTINE convrt(ra, dec, ctoe, az, el)
real(double), INTENT(IN)   :: ra  ! (radians)
real(double), INTENT(IN)   :: dec ! (radians)
real(double), INTENT(IN)   :: ctoe(3, 3)  ! conversion matrix e.g. from CONGEN
real(double), INTENT(OUT)  :: az ! (radians)
real(double), INTENT(OUT)  :: el ! (radians)
!-
real(double) :: cdec, v1, v2, v3, x1, x2, x3
cdec = COS(dec)
v1 = COS(ra) * cdec
v2 = SIN(ra) * cdec
v3 = SIN(dec)
x1 = v1 * ctoe(1,1) + v2 * ctoe(2,1) + v3 * ctoe(3,1)
x2 = v1 * ctoe(1,2) + v2 * ctoe(2,2) + v3 * ctoe(3,2)
x3 = v1 * ctoe(1,3) + v2 * ctoe(2,3) + v3 * ctoe(3,3)
az = ATAN2(x2, x1)
IF(az < 0.0_double) az = az + TWOPI
el = ASIN(x3)
END SUBROUTINE convrt
!---------------------------------------------------------------------
! coord_trans: rotates coordinates to a new frame, all angles in RADIANS
! algorithm from Greisen & Calabretta, 1993 ,AAS Meeting, 182, #09.01
subroutine coord_trans(ra, dec, rapole, decpole, phipole, phi, theta)
real(double),intent(in) :: ra, dec                   ! input coordinates, radians
real(double),intent(in) :: rapole, decpole, phipole  ! New pole and axes
real(double),intent(out) :: phi, theta               ! transformed long, lat, radians
!
real(double) :: radiff, phid, arg1, arg2
!
radiff = ra - rapole
theta  = asin( sin(dec) * sin(decpole) + &
               cos(dec) * cos(decpole) * cos(radiff) )
arg1 = sin(dec) * cos(decpole) - &
       cos(dec) * sin(decpole) * cos(radiff)
arg2 = -cos(dec) * sin(radiff)
phid  = atan2(arg2, arg1) + phipole
phi   = mod(phid+twopi,twopi)
end subroutine coord_trans
!----------------------------------------------------------------------------
! dinvnorm: Inverts cumulative Normal (Gaussian) distribution:
! given a probability  P (area under upper-tail) it returns the normal deviate.
! e.g. dinvnorm(0.5) => 0.0, dinvnorm(0.001) => +3.0902, 
!      dinvnorm(0.999) => -3.0902 etc.
! Note: if p <= 0.0 or >= 1.0 function returns NaN.
! Fortran77 by Peter J Acklam, http://home.online.no/~pjacklam/notes/invnorm
! after routine written by john herrero, ren-raw chen, Rutgers Business School
! Converted using TO_F90 of Alan Miller, restructured by Clive Page, 2009 Oct 3
!
real(double) function dinvnorm(p)
! integer, parameter   :: dp = selected_real_kind(14,50)
real(double), INTENT(IN) :: p ! Probability = upper-tail area. p > 0.0 and p < 1.0
!
real(double), parameter :: &
 a1=-39.6968302866538_double,    a2=220.946098424521_double,   a3=-275.928510446969_double, &
 a4=138.357751867269_double,     a5=-30.6647980661472_double,  a6=2.50662827745924_double, &
 b1=-54.4760987982241_double,    b2=161.585836858041_double,   b3=-155.698979859887_double, &
 b4=66.8013118877197_double,     b5=-13.2806815528857_double, &
 c1=-0.00778489400243029_double, c2=-0.322396458041136_double, c3=-2.40075827716184_double, &
 c4=-2.54973253934373_double,    c5=4.37466414146497_double,   c6=2.93816398269878_double, &
 d1=0.00778469570904146_double,  d2=0.32246712907004_double,   d3=2.445134137143_double, &
 d4=3.75440866190742_double
real(double) :: p_low, p_high, z, q, r
!
p_low  = 0.02425_double
p_high = 1.0_double - p_low
IF(p < p_low) then
   q = SQRT(-2.0_double * LOG(p))
   z = (((((c1*q+c2)*q+c3)*q+c4)*q+c5)*q+c6) / ((((d1*q+d2)*q+d3)*q+d4)*q+1)
else if(p >= p_low .and. p <= p_high) then
   q = p - 0.5_double
   r = q**2
   z = (((((a1*r+a2)*r+a3)*r+a4)*r+a5)*r+a6)*q / &
       (((((b1*r+b2)*r+b3)*r+b4)*r+b5)*r+1.0_double)
else
   q = SQRT(-2.0_double * LOG(1-p))
   z = -(((((c1*q+c2)*q+c3)*q+c4)*q+c5)*q+c6)/((((d1*q+d2)*q+d3)*q+d4)*q+1.0_double)
end if
dinvnorm = -z
END FUNCTION dinvnorm
!----------------------------------------------------------------------------
! erfcc: Returns approximation to Error Function, 
! Note: will be redundant when Fortran2008 arrives, as this has erfc intrinsic
real function erfcc(x)
real, intent(in) :: x
real :: z, t
!
z = abs(x)
t = 1.0/(1.0 + 0.5*z)
erfcc = t * exp(-z*z - 1.26551223 + t * (1.00002368 + t * (0.37409196 + &
    t * (0.09678418 + t * (-0.18628806 + t * (0.27886807 + t * (-1.13520398 + &
    t * (1.48851587 + t * (-0.82215223 + t * 0.17087277)))))))))
if(x < 0.0) erfcc = 2.0 - erfcc
end function erfcc
!----------------------------------------------------------------------------
! find_last_record: number of a direct-access (unformatted) file
subroutine find_last_record(unit, lastrec)
integer, intent(in)  :: unit    ! Number of open direct-access unformatted file
integer, intent(out) :: lastrec ! Last valid record number.
!
integer :: lrec, status, krec, midrec
character(len=1) :: dummy
!
lrec = 1
do
   read(unit, rec=lrec, iostat=status)	dummy
   if(status /= 0) EXIT
   lrec = 2 * lrec
end do
!
! Here LREC is a record beyond the end of the file, but LREC/2 is valid
! Except for the special case when LREC=1, in which case KREC will be set to 0
!
krec = lrec/2
do
   if(krec+1 >= lrec) EXIT
   midrec = (krec + lrec) / 2
   read(unit, rec=midrec, iostat=status) dummy
   if(status /= 0) then
      lrec = midrec
   else
      krec = midrec
   end if
end do
lastrec = krec
end subroutine find_last_record
!----------------------------------------------------------------------------
! gcdist: Great-circle distance between 2 points on celestial sphere, RADIANS
real(double) function gcdistd(ra1, dec1, ra2, dec2)
real(double), intent(in) :: ra1, dec1, ra2, dec2  ! coordinates in radians, result rads
!
  gcdistd = ahav( hav(dec1-dec2) + cos(dec1)*cos(dec2)*hav(ra1-ra2) )
end function gcdistd
!----------------------------------------------------------------------------
real function gcdistr(ra1, dec1, ra2, dec2)
real, intent(in) :: ra1, dec1, ra2, dec2  ! coordinates in radians
!
  gcdistr = ahav( hav(dec1-dec2) + cos(dec1)*cos(dec2)*hav(ra1-ra2) )
end function gcdistr
!-----------------------------------------------------------------------
! ! Returns number of an unused and usable I/O unit.
! subroutine getFreeIoUnit(nunit)
! integer, intent(out) :: nunit   ! unit number free for use
! !-
! logical :: exists, open
! do nunit = 99, 7, -1
!     inquire(unit=nunit, exist=exists, opened=open)
!     if(exists .and. .not. open) return
! end do
! call error('noIOunits', 'no free I/O units left')
! return
! end subroutine getFreeIoUnit
!-------------------------------------------------------------------------------------
! hav: haversine function, angle in RADIANS
real(double) function havd(theta)      
real(double), intent(in) :: theta      ! angle in radians
  havd = (sin(0.5d0*theta))**2
end function havd
!-------------------------------------------------------------------------------------
real function havr(theta)       ! haversign
real, intent(in) :: theta      ! angle in radians
  havr = (sin(0.5*theta))**2
end function havr
!----------------------------------------------------------------------------
! MJD: returns the Modified Julian Date for a given calendar date
integer function mjd(year, month, day)
integer, intent(in) :: year, month, day
INTEGER :: mm
mm   = (month-14)/12          ! mm is -1 for Jan, Feb, zero for other months
mjd  = day - 2432076 + 1461*(year + 4800 + mm)/4 + 367 * &
       (month - 2 - 12*mm) / 12 - 3*((year + 4900 + mm)/100)/4
end function mjd
!-----------------------------------------------------------------------------
! poisson_ge: probability of getting N or more counts when expection is X
real function poisson_ge(x, n)
real, intent(in)    :: x  ! Expected mean
integer, intent(in) :: n  ! Number of counts
!
integer :: j
real(double) :: prob, pge
!
pge  = 1.0d0
prob = exp(-x)
do j = 1, n
   pge = pge - prob
! update probability for term for (j+1)
   prob = prob * x / real(j)
end do
poisson_ge = max(0.0, real(pge))
end function poisson_ge
!-----------------------------------------------------------------------------
! poisson_level: Returns count N such that poisson_ge(x,n) is less than some 
!                specified probability
integer function poisson_level(x, pthr)
real, intent(in) :: x    ! expected or mean level of the distribution
real, intent(in) :: pthr ! Probability threshold, range 0 to 1.0
!
integer :: n
real    :: gauss_dev
!
if(x < 30.0) then
   do n = 0, 200         ! long-stop, should not get here
      if(poisson_ge(x, n) < pthr) EXIT
   end do
   if(n >= 200) then
      call warning('poissonLevel', 'Poisson_level count too large')
   end if
   poisson_level = n
else
   gauss_dev = 0.5 * erfcc(pthr/sqrt(2.0))
   poisson_level = ceiling(x + gauss_dev * sqrt(x))
end if
end function poisson_level
!-----------------------------------------------------------------------------
! poisson_prob: Returns probability of getting exactly N events from a poission
! distribution.  returns p = exp(-x) * x**n / factorial(n)
real function poisson_prob(x, n)
real, intent(in)    :: x  ! Expected mean
integer, intent(in) :: n  ! Number of counts
! Compute this in a loop to get a bit more precision:
integer :: j
poisson_prob = exp(-x)
do j = 1,n
   poisson_prob = poisson_prob * x / j
end do
end function poisson_prob
!-----------------------------------------------------------------------------
! poisson_count: returns event count which has probability < specified value
integer function poisson_count(x, prob)
real, intent(in) :: x    ! Expected mean count level
real, intent(in) :: prob ! Probability level
integer :: n
real    :: p
!
do n = ceiling(x), 100
   p = poisson_prob(x, n)
   if(p < prob) then
      poisson_count = n
      RETURN
   end if
end do
call warning('poissonCount', 'Poisson probability too low')
poisson_count = huge(0)
end function poisson_count
!----------------------------------------------------------------------------
! ran_gauss: returns pair of random deviates with Gaussian distribution
! Uses the Box-Muller transform, see Wikipedia for details.
subroutine ran_gauss(z)
real, intent(out) :: z(2) ! Pair of random values, with mean=0 and sigma=1
!
real :: u(2), f1, f2
!
call random_number(u)
f1 = sqrt(-2.0 * log(u(1)))
f2 = twopi * u(2)
z(1) = f1 * cos(f2)
z(2) = f1 * sin(f2)
end subroutine ran_gauss
!----------------------------------------------------------------------------
! rot_shift: rotates and shifts a celestial position (all coords RADIANS)
subroutine rot_shift(ra0, dec0, param, ra1, dec1, ra2, dec2)
real(double), intent(in) :: ra0, dec0  ! centre of field (radians)
real(double), intent(in) :: param(3)   ! RA, DEC shifts, clockwise rotation angle
real(double), intent(in) :: ra1, dec1  ! point to be rotated
real(double), intent(out) :: ra2, dec2 ! rotated version of ra1,dec1
!
real(double) :: phi, theta
!
call coord_trans(ra1-param(1), dec1-param(2), ra0, dec0, pi, phi, theta)
call coord_trans(phi+param(3), theta, pi, dec0, ra0, ra2, dec2)
end subroutine rot_shift
!----------------------------------------------------------------------------
! rotate: routine to rotate position relative to centre of FOV
subroutine rotate(ra0, dec0, dtheta, ra1, dec1, ra2, dec2)
real(double), intent(in) :: ra0, dec0  ! centre of field (radians)
real(double), intent(in) :: dtheta     ! clockwise rotation angle (radians)
real(double), intent(in) :: ra1, dec1  ! point to be rotated
real(double), intent(out) :: ra2, dec2 ! rotated version of ra1,dec1 (radians)
!
real(double) :: phi, theta
!
call coord_trans(ra1, dec1, ra0, dec0, pi, phi, theta)
phi = phi - dtheta
call coord_trans(phi, theta, pi, dec0, ra0, ra2, dec2)
end subroutine rotate
!------------------------------------------------------------------------------
! splitString: splits a string into array of substrings using specified divider
subroutine splitString(textline, divider, sarray)
character(len=*), intent(in)  :: textline   ! Any line of text
character(len=1), intent(in)  :: divider    ! Dividing character e.g. slash or comma
character(len=*), intent(out) :: sarray(:)  ! Array of string items
! Note if the length of this array is too small, remaining items will be ignored
!
integer :: k, lastdiv, maxitem, item, last
!
lastdiv = 0
maxitem = size(sarray)
do item = 1, maxitem
   k = index(textline(lastdiv+1:), divider)
! when Excel exports as CSV, if a field contains a comma it encloses it in a
! pair of double quotes, these are discarded here.
   if(k > 1) then
      if(textline(lastdiv+1:lastdiv+1) == '"' .and. &
         textline(lastdiv+k-1:lastdiv+k-1) == '"') then
! discard enclosing quotes
            sarray(item) = textline(lastdiv+2:lastdiv+k-2)
      else
            sarray(item) = textline(lastdiv+1:lastdiv+k-1)
      end if
   else if(k == 1) then
      sarray(item) = ' '           ! null argument, replace by blanks
   else
      last = len_trim(textline)
      if(textline(lastdiv+1:lastdiv+1) == '"' .and. textline(last:last) == '"') then
! discard enclosing quotes
         sarray(item) = textline(lastdiv+2:last-1)    ! last argument, rest blank
      else
         sarray(item) = textline(lastdiv+1:last)
      end if
      EXIT
   end if
   lastdiv = lastdiv + k
end do
if(item < maxitem) then
   sarray(item+1:maxitem) = ' '      ! array assignment to set remaining args to blanks
end if
end subroutine splitString
!----------------------------------------------------------------------------
! stats: returns mean and standard deviation of values in an array
subroutine stats(array, avg, sd)
real(double), intent(in) :: array(:)  ! Any double precision array
real(double), intent(out) :: avg, sd  ! Arithmetic mean and standard deviation
!
integer  :: npts
real(double) :: sumsq
!
npts  = size(array)
avg   = sum(array)/npts
sumsq = sum( (array-avg)**2 )
sd    = sqrt(sumsq/max(1,npts-1))
end subroutine stats
end module catcorr_minor
