! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
! catprep_aux.f90
! 2012 Aug 21 v0.3.4 Fixed bug in calculating deltara in twomass_recs, usno_recs
! 2012 Nov  5 v0.3.5 Fixed bug which made USNO access via seq scan in many cases
! 2012 Nov 23 v0.3.6 Detect SDSS version and identify in cat extract file#
! 2012 Dec  3 v0.3.7 Change 2MASS RMAG values to be 2mass k magnitude.
! 2012 Dec 12 v0.3.8 Change identifier of "USNO-B1.0" back to "USNO" as omatt objects.
module catprep_aux
use types
use errorHandling
use ssc_misc	! in ssclib, for getFreeIoUnit()
use constants	! in ssclib, for Pi and TwoPi
use catcorr_minor
implicit none
private
public catprep_top
public fits_status
logical, parameter      :: diag = .false.     ! Set .true. for diagnostic output file
integer, parameter      :: logunit = 9        ! I/O unit for log output
character(*), parameter :: VERSION = "0.4"  ! Written to cat extract file
                                              ! Should be same as SAS version 
interface get_fits_keyword
   module procedure get_fits_keyword_c, get_fits_keyword_d, &
    get_fits_keyword_e, get_fits_keyword_j, get_fits_keyword_l
end interface
real(double), parameter :: HALFPI = 0.5_double * PI
real(double), parameter :: DTOR = PI/180.0_double, RTOD = 180.0_double/PI
real(double), parameter :: ASTOR = pi/(180.0_double * 3600.0_double), &
                           RTOAS = 1.0_double/ASTOR
integer, parameter  :: MAXSRC = 5000   ! Max number of X-ray detections
integer, parameter  :: MAXOPT = 100000 ! Max objects in any optical (or IR) catalogue
integer, parameter  :: MAXCAT = 3      ! Number of reference catalogues
real(double), parameter :: CATRADIUS = 14.0 * 60.0 * ASTOR  ! radius of field of view (rads)
!
!character(*), parameter :: &      !! location of files on Leicester SSC system
!   usnoPath    = '/data/pcms/archive/catalogue/usnob1/binary/', &
!   twomassPath = '/data/pcms/archive/catalogue/2Mass/', &
!   sdssPath    = '/data/pcms/archive/catalogue/SDSS/'

type catType          ! Extracted objects from optical/IR catalogues
   integer  :: nobj = 0         ! Number of catalogue objects in these arrays:
   real(double) :: ra(MAXOPT) = 0.0_double  ! RA, J2000 (degs) 
   real(double) :: dec(MAXOPT) = 0.0_double ! Dec, J2000 (degs)
   real     :: perr(MAXOPT) = 0.0   ! Std error on source postion (degs)
   real     :: bmag(MAXOPT) = 0.0   ! Mean blue magnitude
   real     :: rmag(MAXOPT) = 0.0   ! Mean red magnitude
   integer  :: pmra(MAXOPT) = 0     ! Proper motion in RA, raw units, 0.002 arcsec/yr
   integer  :: pmdec(MAXOPT) = 0    ! ditto, DEC (note used in USNO-B only)
   character(10) :: name = ' '       ! Name of catalogue
end type catType

integer, parameter :: MAXCP = 100    ! Max opt/IR counterparts per X-ray detn
!
CONTAINS
!--------------------------------------------------------------------------
subroutine catprep_top(datafile, outfile, status)
character(*), intent(in) :: datafile  ! Name of an OBSMLI file
character(*), intent(in) :: outfile   ! Name of a FITS file for output
integer, intent(inout)   :: status    ! Inherited status
!
type(catType)   :: cat(3)       ! Storage for up to 3 catalogue extracts
real(double) :: ra_pnt, dec_pnt
real         :: year
character    :: date*10, time*10, logfile*100
integer      :: kslash
!

if(status /= 0) RETURN
if(diag) then
   kslash = index(datafile, '/', .true.)
   logfile = datafile(1:kslash) // 'catprep.log'
   open(unit=logunit, file=logfile, status='replace')
   call date_and_time(date, time)
   write(*, '(5a)') 'catprep ', version, ' run ', date, time
   write(logunit, '(5a)') 'catprep ', version, ' run ', date, time
   write(logunit, '(4a)') 'datafile=', trim(datafile), ' outfile=', trim(outfile)
end if
call read_pointing(datafile, ra_pnt, dec_pnt, year, status)
call cat_read(ra_pnt, dec_pnt, year, cat, status)
call cat_write_extracts(cat, outfile, status)
if(diag) write(*,*)'Logfile: ', trim(logfile), status
end subroutine catprep_top
!-----------------------------------------------------------------------------
subroutine cat_read(cra, cdec, year, cat, status)
! Reads extracts from USNO-B1, 2MASS, and SDSS into data structure cat
real(double), intent(inout)       :: cra, cdec ! Centre of Field of View (radians)
real,     intent(inout)       :: year      ! Year of observation (calendar year)
type(catType), intent(inout) :: cat(:)    ! Returns set of up to 3 catalogue extracts
integer, intent(inout)     :: status    ! Inherited status
!
real(double), parameter :: field_radius = 17.0 * 60.0 * ASTOR
character(len=256) :: ub1_path, twomass_path, sdss_path
!
if(status /= 0) RETURN
! Note new ordering of catalogues in the CAT structure
! Now gets catalogue paths via environment variables
call get_environment_variable('UB1_PATH', ub1_path)
call get_environment_variable('TWOMASS_PATH', twomass_path)
call get_environment_variable('SDSS_PATH',    sdss_path)
call append_slash(ub1_path)
call append_slash(twomass_path)
call append_slash(sdss_path)
call usno_read(ub1_path, cra, cdec, field_radius, year, cat(1), status)
call twomass_read(twomass_path, cra, cdec, field_radius, cat(2), status)
call sdss_read_tiles(sdss_path, cra, cdec, field_radius, cat(3), status)
CONTAINS
! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
subroutine append_slash(string)
character(*), intent(inout) :: string  ! string returned with slash at end 
                                       ! unless it already has one
integer :: length
length = len_trim(string)
if(length == 0) RETURN
if(string(length:length) /= "/") then
   string(length+1:length+1) = "/"
end if                                       
end subroutine append_slash                                       
end subroutine cat_read
!-----------------------------------------------------------------------------
subroutine cat_write_extracts(cat, filename, status)
type(catType), intent(in) :: cat(:)  ! Structure holding catalogue data (3 sets)
character(*), intent(in) :: filename ! Name of FITS file to write
integer, intent(inout)   :: status   ! Inherited status
!
integer, parameter :: ncols = 7
character(16) :: tname(ncols), tform(ncols), tunit(ncols)
data tname / 'CAT_RA','CAT_DEC','CAT_RADEC_ERR','BMAG','RMAG','PMRA',  'PMDEC' /
data tform / '1D',    '1D',     '1E',           '1E',  '1E',  '1J',    '1J' /
data tunit / 'deg',   'deg',    'deg',          'mag', 'mag','2mas/yr','2mas/yr' /
integer :: unit, icat, nrows
logical :: exists
!
if(status /= 0) RETURN
!
! If file already exists delete it
! Note: using portable Fortran way: open, then close(status='delete')
!
inquire(file=filename, exist=exists)
if(exists) then
   call ftgiou(unit, status)
   open(unit=unit, file=filename, access='direct', recl=2880)
   close(unit=unit, status='delete')
   call ftfiou(unit, status)
end if
!
! Create FITS file with empty primary HDU
!
call ftgiou(unit, status)
call ftinit(unit, filename, 10, status)
if(status /= 0) then
   call error('catWriteExtracts', 'error creating file '//trim(filename))
   RETURN
end if
call ftiimg(unit, 8, 0, 0, status)
call ftpkys(unit, 'VERSION', VERSION, 'of catprep', status)
!
! For each catalogue extract with >0 rows, create new HDU and write its data
!
do icat = 1,3
   nrows = cat(icat)%nobj
   if(nrows == 0) CYCLE
   call ftibin(unit, nrows, ncols, tname, tform, tunit, cat(icat)%name, &
               0, status)
   call ftpcld(unit, 1, 1, 1, nrows, cat(icat)%ra(1:nrows),   status)
   call ftpcld(unit, 2, 1, 1, nrows, cat(icat)%dec(1:nrows),  status)
   call ftpcle(unit, 3, 1, 1, nrows, cat(icat)%perr(1:nrows), status)
   call ftpcle(unit, 4, 1, 1, nrows, cat(icat)%bmag(1:nrows), status)
   call ftpcle(unit, 5, 1, 1, nrows, cat(icat)%rmag(1:nrows), status)
   call ftpclj(unit, 6, 1, 1, nrows, cat(icat)%pmra(1:nrows), status)
   call ftpclj(unit, 7, 1, 1, nrows, cat(icat)%pmdec(1:nrows),status)
end do
!
! Close output file which flushes buffers, and release FITS I/O unit
!
call ftclos(unit, status)
call ftfiou(unit, status)
call fits_status('cat_write_extracts', status)
end subroutine cat_write_extracts
!------------------------------------------------------------------------------
!+circle_around finds coordinates of NPTS points around a given point
subroutine circle_points(ra, dec, radius, rapt, decpt)
real(double), intent(in)  :: ra, dec, radius   ! (ra,dec) of centre, radius [radians]
real(double), intent(out) :: rapt(:), decpt(:) ! ra,dec of points (each size NPTS)
! in circle [radians] note wrapped so rapt() is close to ra even if -ve or
! over 2*pi radians in size, for convenience in finding map to use.
!
real(double) :: vz(3), vx(3), ctos(3,3), stoc(3,3), az, el, rapoint, decpoint
integer :: j, npts
!
! Get matrix to a new frame with this position at its centre, meridian arbitrary
!
npts = size(rapt)
call cona2v(ra, dec, vz)
call cona2v(ra, dec-0.5_double, vx)
call congen(vz, vx, ctos)
call conmin(ctos, stoc)
! Generate N points around a circle offset by radius from the pole
do j = 1,npts
   az = (twopi/npts) * j
   el = halfpi - radius
   call convrt(az, el, stoc, rapoint, decpoint)
! Wrap the RA value so that it is near the centre RA even if -ve or over 2pi
   rapt(j)  = mod(rapoint-ra + pi + twopi, twopi) + ra - pi
   decpt(j) = decpoint
end do
end subroutine circle_points
!-----------------------------------------------------------------------------
! Searches header of current HDU for a keyword and returns value
subroutine get_fits_keyword_c(unit, name, value, status)
integer, intent(in)       :: unit   ! I/O unit of open FITS file
character(*), intent(in)  :: name   ! Name of keyword in upper case
character(*), intent(out) :: value  ! Value as string
integer, intent(inout)    :: status ! Inherited status
!
character(50) :: comment ! unused here
if(status /= 0) return
call ftgkys(unit, name, value, comment, status)
if(status /= 0) then
   call error('missingKeyword', 'FITS keyword missing: '//name)
end if
end subroutine get_fits_keyword_c
!-----------------------------------------------------------------------------
! Searches header of current HDU for a keyword and returns value
subroutine get_fits_keyword_d(unit, name, value, status)
integer, intent(in)       :: unit   ! I/O unit of open FITS file
character(*), intent(in)  :: name   ! Name of keyword in upper case
real(double), intent(out) :: value  ! Value as string
integer, intent(inout)    :: status ! Inherited status
!
character(50) :: comment ! unused here
if(status /= 0) return
call ftgkyd(unit, name, value, comment, status)
if(status /= 0) then
   call error('missingKeyword', 'FITS keyword missing: '//name)
end if
end subroutine get_fits_keyword_d
!-----------------------------------------------------------------------------
! Searches header of current HDU for a keyword and returns value
subroutine get_fits_keyword_e(unit, name, value, status)
integer, intent(in)       :: unit   ! I/O unit of open FITS file
character(*), intent(in)  :: name   ! Name of keyword in upper case
real, intent(out)         :: value  ! Value as string
integer, intent(inout)    :: status ! Inherited status
!
character(50) :: comment ! unused here
if(status /= 0) return
call ftgkye(unit, name, value, comment, status)
if(status /= 0) then
   call error('missingKeyword', 'FITS keyword missing: '//name)
end if
end subroutine get_fits_keyword_e
!-----------------------------------------------------------------------------
! Searches header of current HDU for a keyword and returns value
subroutine get_fits_keyword_j(unit, name, value, status)
integer, intent(in)       :: unit   ! I/O unit of open FITS file
character(*), intent(in)  :: name   ! Name of keyword in upper case
integer, intent(out)      :: value  ! Value as string
integer, intent(inout)    :: status ! Inherited status
!
character(50) :: comment ! unused here
if(status /= 0) return
call ftgkyj(unit, name, value, comment, status)
if(status /= 0) then
   call error('missingKeyword', 'FITS keyword missing: '//name)
end if
end subroutine get_fits_keyword_j
!-----------------------------------------------------------------------------
! Searches header of current HDU for a keyword and returns value
subroutine get_fits_keyword_l(unit, name, value, status)
integer, intent(in)       :: unit   ! I/O unit of open FITS file
character(*), intent(in)  :: name   ! Name of keyword in upper case
logical, intent(out)      :: value  ! Value as string
integer, intent(inout)    :: status ! Inherited status
!
character(50) :: comment ! unused here
if(status /= 0) return
call ftgkyl(unit, name, value, comment, status)
if(status /= 0) then
   call error('missingKeyword', 'FITS keyword missing: '//name)
end if
end subroutine get_fits_keyword_l
!-----------------------------------------------------------------------------
subroutine fits_status(subname, status)
! Reports on any non-zero FITS status value with message to standard output
character(*), intent(in) :: subname   ! Name of subroutine or function
integer, intent(in)      :: status    ! FITS status value to check
!
character :: msg*200
!
if(status == 0) RETURN
call ftgerr(status, msg)
write(errStr,'(a,i4,4a)') 'FITSIO error', status, ' in ', trim(subname), &
                     ' - ', trim(msg)
call error('fitsioError')
end subroutine fits_status
!-----------------------------------------------------------------------------
subroutine read_pointing(OBSMLIfile, ra_pnt, dec_pnt, year, status)
! Reads pointing axis and year of observation from a data product
character(*), intent(in)      :: OBSMLIfile      ! FITS file name of type OBSMLI
real(double), intent(inout)     :: ra_pnt          ! Pointing axis RA [radians]
real(double), intent(inout)     :: dec_pnt         ! Pointing axis DEC [radians]
real, intent(inout)             :: year            ! Year of observation [years]
integer, intent(inout)        :: status          ! Inherited status
!
integer   :: iyear, month, date, unit
character :: date_obs*30
real      :: secs
!
if(status /= 0) RETURN
!
! Here open file READONLY
!
call ftgiou(unit, status)
call ftnopn(unit, OBSMLIfile, 0, status) 
call fits_status('read_pointing/ftnopn', status)
if(status /= 0) RETURN
!
! Get following scalars from FITS primary header
!
call get_fits_keyword(unit, 'RA_PNT', ra_pnt, status)
call get_fits_keyword(unit, 'DEC_PNT', dec_pnt, status)
call get_fits_keyword(unit, 'DATE-OBS', date_obs, status)
! Convert pointing axis coordinates from degrees to radians
ra_pnt = ra_pnt * dtor
dec_pnt = dec_pnt * dtor
if(status /= 0) go to 1000
!
! date of observation needed for proper-motion calculation, convert to real year
!
read(date_obs, '(i4,1x,i2,1x,i2)', iostat=status) iyear, month, date
if(status == 0) then
   year = 2000.0 + (mjd(iyear, month, date) - mjd(2000, 1, 1)) / 365.25
else
   call warning('invalidDateObs', 'error decoding DATE-OBS value')
   year = 2000.0
end if
1000 CONTINUE
call ftclos(unit, status)
call ftfiou(unit, status)
call fits_status('read_pointing', status)
if(diag) then
   call cpu_time(secs)
   write(LOGUNIT,*) 'read_pointing', ra_pnt*rtod, dec_pnt*rtod, status, secs
end if
end subroutine read_pointing
!------------------------------------------------------------------------------
subroutine sdss_version(path, sdssid)
character(*), intent(in)  :: path   ! Path to SDSS csv files with trailing slash
character(*), intent(out) :: sdssid ! Identifier e.g. "SDSS-DR9"
!
! The SDSS files do not include any version identifier.  With the benefit of
! hindsight this should have been done.  This routine is a stopgap to work out
! which version of SDSS is being accessed.
!
integer :: unit, status
real(double) :: ra
!
sdssid = 'SDSS-?'                  ! Default value as fall-back
call getFreeIOunit(unit)
open(unit=unit, file=trim(path)//'sdss_100_005.csv', status='old', &
     iostat=status)
if(status /= 0) RETURN
read(unit, '()')            ! skip first record which has header
read(unit, *) ra
if(abs(ra - 100.458053) < 0.001) then
   sdssid = 'SDSS-DR8'
else if(abs(ra - 100.024578) < 0.001) then
   sdssid = 'SDSS-DR9'
end if
close(unit)
end subroutine sdss_version
!------------------------------------------------------------------------------
subroutine sdss_read_tiles(path, ra, dec, radius, cat, status)
character(*), intent(in) :: path    ! Path to SDSS csv files with trailing slash
real(double), intent(in) :: ra, dec ! Centre of field [radians]
real(double), intent(in) :: radius  ! radius of field [radians]
type(catType), intent(out) :: cat   ! Catalogue extract structure
integer, intent(inout)   :: status  ! inherited status
!
integer, parameter :: npts = 50       ! No of points around circle to store
real(double) :: rapt(npts), decpt(npts)   ! stores ra and dec of points on circle
!
integer :: miny, maxy, j, ix, iy, kx, nlines, nrecs, unit, objtype
integer, allocatable :: minx(:), maxx(:)
character :: filename*250, title*1, sdssid*10
real(double) :: sra, sdec, poserr, dist
real :: mag(5), secs
!
! Get a list of points on the circle of given radius around (ra,dec)
!
if(status /= 0) RETURN
cat%nobj = 0
! At low declinations no point in searching, no sky coverage
if(dec*RTOD < -26.0) RETURN
call circle_points(ra, dec, radius, rapt, decpt)
!
! Find minimum and maximum y value for tiles
!
miny = minval(floor(90.0 - decpt*RTOD))
maxy = maxval(floor(90.0 - decpt*RTOD))
allocate(minx(miny:maxy), maxx(miny:maxy))
minx = 1000  ! note an array assignment to out-of-band value
maxx = -1000
do j = 1,npts
   iy = floor(90.0 - decpt(j)*RTOD)
   ix = floor(rapt(j)*rtod)            ! value may be -ve or >= 360
   minx(iy) = min(ix, minx(iy))
   maxx(iy) = max(ix, maxx(iy))
end do
!
nrecs = 0
do iy = miny, maxy
   do ix = minx(iy), maxx(iy)
      kx = mod(ix+360,360)
      write(filename, "(2a,i3.3,a,i3.3,a)") trim(path), 'sdss_', &
        kx, '_', iy, '.csv'
      call getFreeIoUnit(unit)
      open(unit=unit, file=filename, status='old', iostat=status)
!  Ignore file not found message - may be a part of sky not covered by SDSS
      if(status /= 0) then
         status = 0
         CYCLE
      end if
      read(unit,'(a)') title
      nlines = 0
      do 
         read(unit, *, iostat=status) sra, sdec, poserr, objtype, mag
         if(status /= 0) then
            status = 0
            EXIT
         end if
! Note objtype is 3 or 6, currently unused here, and only 2 magnitudes used
         nlines = nlines + 1
         dist = gcdist(sra*DTOR, sdec*DTOR, ra, dec)
         if(dist < radius) then
            if(nrecs >= MAXOPT) stop 'too many SDSS records to store'
            nrecs = nrecs + 1
            cat%ra(nrecs)   = sra 
            cat%dec(nrecs)  = sdec 
            cat%perr(nrecs) = poserr / 3600.0  ! convert to degrees
            cat%bmag(nrecs) = mag(1)           ! SDSS 'u' magnitude
            cat%rmag(nrecs) = mag(3)           ! SDSS 'r' magnitude
         end if
      end do
      close(unit=unit)
   end do
end do
cat%nobj = nrecs
call sdss_version(path, sdssid)
cat%name = sdssid
if(diag) then
   call cpu_time(secs)
   write(LOGUNIT,*) trim(filename), status, secs
end if
status = 0
end subroutine sdss_read_tiles
!-----------------------------------------------------------------------------
! twomass_read: does cone-search on 2MASS binary files
subroutine twomass_read(twomasspath, cra, cdec, Radius, cat, status)
character(*), intent(in) :: twomassPath ! twomass catalogue path with trailing slash
real(double), intent(in)   :: cra, cdec ! Centre of Field of View, radians
real(double), intent(in)   :: radius    ! Cone radius, radians
type(catType), intent(out) :: cat   ! Catalogue extract
integer, intent(inout) :: status    ! inherited status
!
integer :: ncat, nread, minzone, maxzone, zone, unit, nrecs, nzone, nzok, &
           iset, nsets, minrec(2), maxrec(2), irec
real(double) :: decmin, decmax, ra, dec, offset
real     :: jmag, hmag, kmag, perr, secs
character :: czone*4, datafile*250
!
if(status /= 0) RETURN
cat%nobj = 0
ncat   = 0
nread  = 0
decmin = cDec - Radius
decmax = cDec + Radius
minzone =   floor(10.0_double * (decmin * rtod + 90.0_double)) 
maxzone = ceiling(10.0_double * (decmax * rtod + 90.0_double)) 
zones: &
do zone = minzone, maxzone
!
! Determine name of binary data file and open it
!
   write(czone, '(i4.4)') zone
   datafile = trim(twomassPath) // '2mass_' // czone // '.bin'
   call getFreeIoUnit(unit)
   open(unit=unit, file=datafile, status='old', access='direct', recl=32, &
        iostat=status)
   if(status /= 0) then
      close(unit)
      call error('twomassRead', 'error opening twomass file '// trim(datafile))
      RETURN
   end if
!
! Find length of file in bytes, divide by 80 to get number of records
!
   read(unit, rec=1) nrecs
   nzone = 0
   nzok  = 0
   call twomass_recs(unit, cra, cdec, radius, nrecs, nsets, &
                     minrec, maxrec, status)
!
! Cope with case of RA range which happens to include 0 = 360 degsrees:
! In this case ramax will be less than ramin, need to do 2 read scans
! 0 to lrec, krec to nrecs
!
   do iset = 1, nsets
      do irec = minrec(iset), maxrec(iset)
         read(unit, rec=irec, iostat=status) ra, dec, perr, jmag, hmag, kmag
         if(status /= 0) then
            close(unit)
            call error('twomassRecord', 'error reading record')
            CYCLE zones     
         end if
         nread = nread + 1
         nzone = nzone + 1
!
! skip this point if it is not within the cone
!
         offset = gcdist(cRa, cDec, ra*dtor, dec*dtor)
         if(offset > radius) CYCLE
         nzok = nzok + 1
         if(ncat >= MAXOPT) then
            call error('twomassSize', 'twomass_read too many data points')
            RETURN
         end if
         ncat = ncat + 1
         cat%ra(ncat)   = ra
         cat%dec(ncat)  = dec
         cat%perr(ncat) = perr / 3600.0  ! Convert arcsecs to degrees
         cat%bmag(ncat) = 0.0  ! infra-red catalog, no blue magnitude here.
         cat%rmag(ncat) = kmag
         cat%nobj       = ncat
      end do    ! irec
   end do       ! irange
   close(unit=unit)
end do zones         ! zone
cat%nobj = ncat
cat%name = "2MASS"
if(diag) then
   call cpu_time(secs)
   write(LOGUNIT,*) trim(datafile), status, secs
end if
status = 0
end subroutine twomass_read
!-----------------------------------------------------------------------
! Computes for each zone file the range of RA values subtended by the cone
subroutine twomass_recs(unit, cra, cdec, radius, nrecs, nsets, &
                     minrec, maxrec, status)
integer, intent(in)  :: unit       ! I/O unit of opened USNO binary file
real(double), intent(in) :: cra, cdec  ! coordinates of centre of cone (radians)
real(double), intent(in) :: radius     ! radius of cone (radians)
integer, intent(in)  :: nrecs      ! No of records in this zone's file
integer, intent(out) :: nsets      ! 1 set normally, 2 if RA range covers 0/360
integer, intent(out) :: minrec(2), maxrec(2) ! record ranges of 1 or 2 sets
integer, intent(inout) :: status      ! Inherited status
!
real(double) :: decmax, deltara, ramin, ramax
integer :: krec, lrec
if(status /= 0) RETURN
!
! Calculation of deltra below is a slight over-estimate, but there is a proper
! cone-search later which will remove the unwanted points
!
decmax = min(abs(cdec) + radius, 1.569_double)
deltara = radius / cos(decmax)
ramin    = mod(twopi + cra - deltara, twopi)
ramax    = mod(twopi + cra + deltara, twopi)
call twomass_search(unit, nrecs, ramin, krec, status)
call twomass_search(unit, nrecs, ramax, lrec, status)
if(ramax >= ramin) then
   nsets = 1
   minrec(1) = krec
   maxrec(1) = lrec
else
   nsets = 2
   minrec(1) = 2
   maxrec(1) = lrec
   minrec(2) = krec
   maxrec(2) = nrecs
end if
end subroutine twomass_recs
!-----------------------------------------------------------------------
! Does binary search of 2MASS file to find record with given RA
subroutine twomass_search(unit, nrecs, ra, krec, status)
integer, intent(in)  :: unit  ! I/O unit number on which file opened
integer, intent(in)  :: nrecs ! Length of the file in records
real(double), intent(in) :: ra    ! RA to search for [RADIANS]
integer, intent(out) :: krec  ! first record with this RA
integer, intent(inout) :: status ! Inherited status
!
integer :: lorec, hirec, midrec
real(double) :: radegs, rafound
!
if(status /= 0) RETURN
radegs = ra * rtod
lorec = 2
hirec = nrecs
do
   midrec = (lorec + hirec) / 2
   read(unit=unit, rec=midrec, iostat=status) rafound
   if(status /= 0) then
      close(unit)
      call error('twomassSearch', 'error finding record')
      RETURN
   end if
   if(rafound < radegs) then
      lorec = midrec
   else
      hirec = midrec
   end if
   if(lorec + 1 >= hirec) EXIT
end do
krec = lorec
end subroutine twomass_search
!-----------------------------------------------------------------------
! Computes mean of blue or red magnitudes ignoring nulls
real function usno_mag(m1, m2, avmag)
real, intent(in) :: m1, m2 ! magnitudes from different surveys
real, intent(in) :: avmag  ! mean of all red/blue magnitudes as default        
!
integer :: n
real    :: sum
sum = 0.0
n   = 0
if(m1 > 0.0 .and. m1 < 30.0) then
   sum = sum + m1
   n   = n + 1
end if
if(m2 > 0.0 .and. m2 < 30.0) then
   sum = sum + m2
   n   = n + 1
end if
! If neither magnitude is value, use the mean of all 4 as default
if(n == 0) then
   usno_mag = avmag
else
   usno_mag = sum / n
end if
end function usno_mag
!-----------------------------------------------------------------------------
! Does cone-search on USNOB and returns data to structure cat.
subroutine usno_read(usnoPath, cra, cdec, radius, year, cat, status)
character(*), intent(in) :: usnoPath ! USNO-B catalogue path with trailing slash
real(double), intent(in) :: cra, cdec ! Centre of Field of View, radians
real(double), intent(in) :: radius    ! Cone radius, radians
real, intent(in)         :: year      ! Year e.g. 2009.3 for proper motion use
type(catType), intent(out) :: cat  ! Catalogue extract
integer, intent(inout) :: status   ! inherited status
!
real(double) :: decmin, decmax, ra, dec, offset
real     :: raerr, decerr, poserr, magsum, magnitude(4), pmra, pmdec, avmag, secs
integer  :: minzone, maxzone, zone, unit, ncat, nread, nrecs, nzone, &
            lenrec, nmags, irec, imag, pmprob, npm, ipmra, ipmdec
integer   :: minrec, maxrec, inbuf(10), dummyrec(20) = 0
character :: czone*4, datafile*250

!
if(status /= 0) RETURN
inquire(iolength=lenrec) dummyrec
cat%nobj = 0
ncat   = 0
nread  = 0
npm    = 0
decmin = cDec - Radius
decmax = cDec + Radius
minzone =   floor(10.0_double * (decmin * rtod + 90.0_double)) 
maxzone = ceiling(10.0_double * (decmax * rtod + 90.0_double)) 
zones: &
do zone = minzone, maxzone
!
! Determine name of binary data file and open it
!
   write(czone, '(i4.4)') zone
   datafile = trim(usnoPath) // trim(czone(1:3)) // trim('/b') // trim(czone) // '.cat'

   call getFreeIoUnit(unit)
   open(unit=unit, file=datafile, status='old', access='direct', recl=lenrec, &
        iostat=status)
   if(diag) write(logunit,*)'Opened ', trim(datafile), status
   if(status /= 0) then
      call error('usnoRead', 'error opening USNO file '// trim(datafile))
      RETURN
   end if
!
! Find last valid record number of the file
!
   call find_last_record(unit, nrecs)
   nzone = 0
   call usno_recs(unit, cra, cdec, radius, nrecs, minrec, maxrec, status)
!   if(diag) write(logunit,*)'usno_recs', cra*rtod, cdec*rtod, radius*rtod,&
!            nrecs, minrec, maxrec 
   do irec = minrec, maxrec
         read(unit, rec=irec, iostat=status) inbuf
         if(status /= 0) then
            close(unit)
            call error('usnoRecord', 'error reading USNO record')
            CYCLE zones     
         end if
         nread = nread + 1
         nzone = nzone + 1
! Store ra,dec in DEGREES
         ra  = inbuf(1) * 0.01_double * ASTOR * RTOD  ! file units are 0.01 arc-secs
         dec = ((inbuf(2) * 0.01_double * ASTOR) - halfpi) * RTOD  ! ditto
!
! skip this point if it is not within the cone
!
         offset = gcdist(cRa, cDec, ra*dtor, dec*dtor)
         if(offset > radius) CYCLE
         if(ncat >= MAXOPT) then
            call error('usnoSize', 'USNO too many data points')
            RETURN
         end if
! Check if Tycho infill source, if so assume posn error is 1 milli-arcsec
         if(inbuf(4) < 100000000) then
            raerr = 0.0
            decerr = 0.0
            poserr = 1.0 / 3600000.0
         else
            raerr  = mod(inbuf(5),1000) * cos(dec)  ! milli-arcseconds
            decerr = mod(inbuf(5)/1000,1000)
            poserr = sqrt(raerr**2 + decerr**2) / 3600000.0
         end if
! 
! Buffer elements 5 and 6 have proper motions, in units of 0.002 arcsec/year
! pmprob has range 0 to 9.  Use them if pmprob is 5 or above.
!
         if(ncat >= MAXOPT) STOP 'too many positions in catalogue to store'
         ncat = ncat + 1
         pmprob = mod(inbuf(3)/100000000, 10)
         if(pmprob > 5) then
            ipmra   = (mod(inbuf(3),10000) - 5000)
            ipmdec  = (mod(inbuf(3)/10000,10000) - 5000)
            cat%pmra(ncat) = ipmra
            cat%pmdec(ncat) = ipmdec
            pmra  = ipmra  * 0.002 * ASTOR * (year - 2000.0)
            pmdec = ipmdec * 0.002 * ASTOR * (year - 2000.0)
            ra = ra + pmra
            dec = dec + pmdec
            if(0.002 * sqrt(real(ipmra**2 + ipmdec**2)) > 0.1) then
               npm = npm + 1
            end if
         end if
         cat%ra(ncat)   = ra
         cat%dec(ncat)  = dec
         cat%perr(ncat) = poserr
         cat%nobj       = ncat
!
! Buffer elements 6 to 9 have blue1, red1, blue2, red2 magnitudes 
! detect null value of 0 and 99.99, otherwise take mean of each of blue and red
!
         nmags = 0
         magsum = 0.0
         do imag = 1,4
            magnitude(imag) = 0.01 * mod(inbuf(imag+5), 10000)
            if(magnitude(imag) > 0.0 .and. magnitude(imag) < 30.0) then
               nmags = nmags + 1
               magsum = magsum + magnitude(imag)
            end if
         end do
         avmag = magsum / max(1,nmags)
         cat%bmag(ncat) = usno_mag(magnitude(1), magnitude(3), avmag)
         cat%rmag(ncat) = usno_mag(magnitude(2), magnitude(4), avmag)
      end do    ! irec
      close(unit=unit)
end do zones         ! zone
cat%nobj = ncat
cat%name = "USNO"
if(diag) then
   call cpu_time(secs)
   write(LOGUNIT,*) trim(datafile), status, secs
end if
status = 0
end subroutine usno_read
!-----------------------------------------------------------------------
! Computes for each zone file the range of RA values subtended by the cone
subroutine usno_recs(unit, cra, cdec, radius, nrecs, &
                     minrec, maxrec, status)
integer, intent(in)  :: unit           ! I/O unit of opened USNO binary file
real(double), intent(in) :: cra, cdec  ! coordinates of centre of cone (radians)
real(double), intent(in) :: radius     ! radius of cone (radians)
integer, intent(in)  :: nrecs          ! No of records in this zone's file
integer, intent(out) :: minrec, maxrec ! record ranges to use in file
integer, intent(inout) :: status       ! Inherited status
!
real(double) :: deltara, ramin, ramax, decmax
!
if(status /= 0) RETURN
!
! Calculation of deltra below is a slight over-estimate, but there is a proper
! cone-search later which will ensure unwanted points are not used
!
decmax = min(abs(cdec) + radius, 1.569_double)
deltara = radius / cos(decmax)
ramin = cra - deltara
ramax = cra + deltara
if(ramin < 0.0_double .or. ramax > 2.0*pi) then
! In the case that the range overlaps zero point of RA, search the whole file 
   minrec = 1
   maxrec = nrecs
else
! Search for the record numbers corresponding to ramin, ramax
   call usno_search(unit, nrecs, ramin, minrec, status)
   call usno_search(unit, nrecs, ramax, maxrec, status)
end if
end subroutine usno_recs
!-----------------------------------------------------------------------
! Does binary search of USNO-B file to find record with given RA
subroutine usno_search(unit, nrecs, ra, krec, status)
integer, intent(in)  :: unit  ! I/O unit number on which file opened
integer, intent(in)  :: nrecs ! Length of the file in records
real(double), intent(in) :: ra    ! RA to search for
integer, intent(out) :: krec  ! first record with this RA
integer, intent(inout) :: status ! Inherited status
!
integer :: lorec, hirec, midrec, inbuf
real(double) :: rafound
!
if(status /= 0) RETURN
lorec = 1
hirec = nrecs
do
   midrec = (lorec + hirec) / 2
   read(unit=unit, rec=midrec, iostat=status) inbuf
   if(status /= 0) then
      close(unit)
      call error('usnoSearch', 'error finding record in USNO')
      RETURN
   end if
   rafound = inbuf * 0.01_double * ASTOR   ! file units 0.01 arc-seconds
   if(rafound < ra) then
      lorec = midrec
   else
      hirec = midrec
   end if
   if(lorec + 1 >= hirec) EXIT
end do
krec = lorec
end subroutine usno_search
!--
end module catprep_aux
